/*************************************************
* DER Coding Source File                         *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/der_enc.h>

namespace Botan {

namespace DER {

/*************************************************
* DER encode NULL                                *
*************************************************/
void encode_null(DER_Encoder& coder)
   {
   coder.add_object(NULL_TAG, 0, 0);
   }

/*************************************************
* DER encode a BOOLEAN                           *
*************************************************/
void encode(DER_Encoder& coder, bool is_true)
   {
   if(is_true)
      coder.add_object(BOOLEAN, 0xFF);
   else
      coder.add_object(BOOLEAN, 0x00);
   }

/*************************************************
* DER encode a small INTEGER                     *
*************************************************/
void encode(DER_Encoder& coder, u32bit n)
   {
   encode(coder, BigInt(n));
   }

/*************************************************
* DER encode an INTEGER                          *
*************************************************/
void encode(DER_Encoder& coder, const BigInt& n)
   {
   if(n == 0)
      coder.add_object(INTEGER, 0);
   else
      {
      bool extra_zero = (n.bits() % 8 == 0);
      SecureVector<byte> contents(extra_zero + n.bytes());
      encode(contents + extra_zero, n);
      if(n < 0)
         {
         for(u32bit j = 0; j != contents.size(); j++)
            contents[j] = ~contents[j];
         for(u32bit j = contents.size(); j > 0; j--)
            if(++contents[j-1])
               break;
         }
      coder.add_object(INTEGER, contents, contents.size());
      }
   }

/*************************************************
* DER encode an OCTET STRING or BIT STRING       *
*************************************************/
void encode(DER_Encoder& coder, const SecureVector<byte>& octets,
            ASN1_Tag tag)
   {
   encode(coder, octets, octets.size(), tag);
   }

/*************************************************
* DER encode an OCTET STRING or BIT STRING       *
*************************************************/
void encode(DER_Encoder& coder, const byte octets[], u32bit length,
            ASN1_Tag tag)
   {
   if(tag != OCTET_STRING && tag != BIT_STRING)
      throw Invalid_Argument("DER_Encoder: Invalid tag for byte string");

   if(tag == OCTET_STRING)
      coder.add_object(OCTET_STRING, octets, length);
   else
      {
      SecureVector<byte> encoded;
      encoded.append(0);
      encoded.append(octets, length);
      coder.add_object(BIT_STRING, encoded, encoded.size());
      }
   }

/*************************************************
* DER encode an OBJECT IDENTIFIER                *
*************************************************/
void encode(DER_Encoder& coder, const OID& oid_obj)
   {
   std::vector<u32bit> oid = oid_obj.get_id();

   SecureVector<byte> encoding;
   encoding.append(40 * oid[0] + oid[1]);

   for(u32bit j = 2; j != oid.size(); j++)
      {
      if(oid[j] == 0)
         encoding.append(0);
      else
         {
         u32bit blocks = high_bit(oid[j]) + 6;
         blocks = (blocks - (blocks % 7)) / 7;

         for(u32bit k = 0; k != blocks - 1; k++)
            encoding.append(0x80 | ((oid[j] >> 7*(blocks-k-1)) & 0x7F));
         encoding.append(oid[j] & 0x7F);
         }
      }
   coder.add_object(OBJECT_ID, encoding, encoding.size());
   }

}

}
