/*************************************************
* DER Encoder Source File                        *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/der_enc.h>
#include <botan/bigint.h>

namespace Botan {

namespace {

/*************************************************
* DER encode an ASN.1 type tag                   *
*************************************************/
SecureVector<byte> encode_tag(ASN1_Tag tag)
   {
   SecureVector<byte> encoded_tag;
   if(tag <= 30)
      encoded_tag.append((byte)tag);
   else
      {
      u32bit blocks = high_bit(tag) + 6;
      blocks = (blocks - (blocks % 7)) / 7;

      for(u32bit k = 0; k != blocks - 1; k++)
         encoded_tag.append(0x80 | ((tag >> 7*(blocks-k-1)) & 0x7F));
      encoded_tag.append(tag & 0x7F);
      }
   return encoded_tag;
   }

/*************************************************
* DER encode an ASN.1 length field               *
*************************************************/
SecureVector<byte> encode_length(u32bit length)
   {
   SecureVector<byte> encoded_length;
   if(length <= 127)
      encoded_length.append((byte)length);
   else
      {
      const u32bit top_byte = significant_bytes(length);
      encoded_length.append((byte)(0x80 | top_byte));
      for(u32bit j = 4-top_byte; j != 4; j++)
         encoded_length.append(get_byte(j, length));
      }
   return encoded_length;
   }

}

/*************************************************
* Return the encoded SEQUENCE                    *
*************************************************/
SecureVector<byte> DER_Encoder::DER_Sequence::get_contents()
   {
   SecureVector<byte> encoded_length = encode_length(contents.size());
   SecureVector<byte> encoded_tag = encode_tag(ASN1_Tag(tag | 0x20));

   SecureVector<byte> retval;
   retval.append(encoded_tag);
   retval.append(encoded_length);
   retval.append(contents);
   contents.destroy();
   return retval;
   }

/*************************************************
* Add an encoded value to the SEQUENCE           *
*************************************************/
void DER_Encoder::DER_Sequence::add_bytes(const byte data[], u32bit length)
   {
   contents.append(data, length);
   }

/*************************************************
* DER_Sequence Constructor                      *
*************************************************/
DER_Encoder::DER_Sequence::DER_Sequence(ASN1_Tag t) : tag(t)
   {
   }

/*************************************************
* Return the encoded contents                    *
*************************************************/
SecureVector<byte> DER_Encoder::get_contents()
   {
   if(sequence_level != 0)
      throw Invalid_State("DER_Encoder: Sequence hasn't been marked done");

   SecureVector<byte> retval;
   retval = contents;
   contents.destroy();
   return retval;
   }

/*************************************************
* Start a new ASN.1 sequence                     *
*************************************************/
void DER_Encoder::start_sequence(ASN1_Tag tag)
   {
   sequence_level++;
   subsequences.push_back(DER_Sequence(tag));
   }

/*************************************************
* Finish the current ASN.1 sequence              *
*************************************************/
void DER_Encoder::end_sequence(ASN1_Tag tag)
   {
   if(sequence_level == 0)
      throw Invalid_State("DER_Encoder::end_sequence: No such sequence");
   if(subsequences.at(sequence_level-1).tag_of() != tag)
      throw Invalid_Argument("DER_Encoder::end_sequence: Tag mismatch");

   SecureVector<byte> seq = subsequences.at(sequence_level-1).get_contents();
   sequence_level--;
   add_raw_octets(seq);
   }

/*************************************************
* Write raw octets into the stream               *
*************************************************/
void DER_Encoder::add_raw_octets(const SecureVector<byte>& octets)
   {
   if(sequence_level == 0)
      contents.append(octets, octets.size());
   else
      subsequences.at(sequence_level-1).add_bytes(octets, octets.size());
   }

/*************************************************
* Write the encoding of the octet(s)             *
*************************************************/
void DER_Encoder::add_object(ASN1_Tag tag, const byte rep[], u32bit length)
   {
   SecureVector<byte> encoded_tag = encode_tag(tag);
   SecureVector<byte> encoded_length = encode_length(length);

   SecureVector<byte> buffer;
   buffer.append(encoded_tag);
   buffer.append(encoded_length);
   buffer.append(rep, length);

   add_raw_octets(buffer);
   }

/*************************************************
* Write the encoding of the octet                *
*************************************************/
void DER_Encoder::add_object(ASN1_Tag tag, byte rep)
   {
   add_object(tag, &rep, 1);
   }

/*************************************************
* DER_Encoder Constructor                        *
*************************************************/
DER_Encoder::DER_Encoder()
    {
   sequence_level = 0;
   }

}
