/*************************************************
* Discrete Logarithm Parameters Source File      *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/dl_param.h>
#include <botan/primes.h>
#include <botan/numthry.h>
#include <botan/asn1.h>
#include <botan/pem.h>

namespace Botan {

/*************************************************
* DL_Group Constructor                           *
*************************************************/
DL_Group::DL_Group()
   {
   group_type = Uninitialized;
   }

/*************************************************
* DL_Group Constructor                           *
*************************************************/
DL_Group::DL_Group(u32bit pbits, PrimeType type)
   {
   if(pbits < 512)
      throw Invalid_Argument("DL_Group: prime size " + to_string(pbits) +
                             " is too small");

   if(type == Strong)
      group_type = DH_Group;
   else
      group_type = DSA_Group;

   if(type == Strong)
      {
      p = random_safe_prime(pbits, PublicValue);
      q = (p - 1) / 2;
      g = 2;
      }
   else if(type == Prime_Subgroup || type == DSA_Kosherizer)
      {
      if(type == Prime_Subgroup)
         {
         const u32bit qbits = 2 * dl_work_factor(pbits);
         q = random_prime(qbits, PublicValue);
         BigInt X;
         while(p.bits() != pbits || !is_prime(p))
            {
            X = random_integer(pbits, PublicValue);
            p = X - (X % (2*q) - 1);
            }
         }
      else
         generate_dsa_primes(p, q, pbits);

      BigInt e = (p - 1) / q;
      u32bit prime_no = 0;
      do
         g = power_mod(PRIMES[prime_no++], e, p);
      while(g == 1);
      }
   }

/*************************************************
* DL_Group Constructor                           *
*************************************************/
DL_Group::DL_Group(const SecureVector<byte>& seed, u32bit pbits, u32bit start)
   {
   if(!generate_dsa_primes(p, q, seed, seed.size(), pbits, start))
      throw Invalid_Argument("DL_Group: The seed/counter given does not "
                             "generate a DSA group");

   BigInt e = (p - 1) / q;
   u32bit prime_no = 0;
   do
      g = power_mod(PRIMES[prime_no++], e, p);
   while(g == 1);
   group_type = DSA_Group;
   if(!fast_check())
      throw Invalid_Argument("DL_Group: Invalid primes have been specified");
   }

/*************************************************
* DL_Group Constructor                           *
*************************************************/
DL_Group::DL_Group(const BigInt& p1, const BigInt& g1)
   {
   if(p1 < 3 || g1 <= 1 || g1 >= p1)
      throw Invalid_Argument("DL parameters: Invalid prime or generator");

   p = p1;
   g = g1;
   q = (p1 - 1) / 2;
   group_type = DH_Group;
   return; // FIXME
   if(!fast_check())
      throw Invalid_Argument("DL_Group: Invalid primes have been specified");
   }

/*************************************************
* DL_Group Constructor                           *
*************************************************/
DL_Group::DL_Group(const BigInt& p1, const BigInt& q1, const BigInt& g1)
   {
   if(p1 < 3 || q1 < 3 || q1 >= p1 || g1 <= 1 || g1 >= p1)
      throw Invalid_Argument("DL parameters: Invalid prime(s) or generator");

   p = p1;
   g = g1;
   q = q1;

   if((p - 1) / 2 == q1)
      group_type = DH_Group;
   else
      group_type = DSA_Group;
   if(!fast_check())
      throw Invalid_Argument("DL_Group: Invalid primes have been specified");
   }

/*************************************************
* Verify the parameters                          *
*************************************************/
bool DL_Group::verify() const
   {
   if(type() == Uninitialized)
      throw Invalid_State("DLP group cannot be used uninitialized");
   if((p - 1) % q != 0)
      return false;
   if(!verify_prime(p) || !verify_prime(q))
      return false;
   return true;
   }

/*************************************************
* Do a fast check of the parameters              *
*************************************************/
bool DL_Group::fast_check() const
   {
   if(type() == Uninitialized)
      throw Invalid_State("DLP group cannot be used uninitialized");
   if((p - 1) % q != 0)
      return false;
   if(!check_prime(p) || !check_prime(q))
      return false;
   if(type() == DH_Group && ((p - 1) / q != 2))
      return false;
   return true;
   }

/*************************************************
* Return the prime                               *
*************************************************/
const BigInt& DL_Group::get_p() const
   {
   if(type() == Uninitialized)
      throw Invalid_State("DLP group cannot be used uninitialized");
   return p;
   }

/*************************************************
* Return the subgroup                            *
*************************************************/
const BigInt& DL_Group::get_q() const
   {
   if(type() == Uninitialized)
      throw Invalid_State("DLP group cannot be used uninitialized");
   return q;
   }

/*************************************************
* Return the generator                           *
*************************************************/
const BigInt& DL_Group::get_g() const
   {
   if(type() == Uninitialized)
      throw Invalid_State("DLP group cannot be used uninitialized");
   return g;
   }

/*************************************************
* See if the prime is st p = 2*q + 1, q prime    *
*************************************************/
DL_Group::GroupType DL_Group::type() const
   {
   return group_type;
   }

/*************************************************
* See if the prime is st p = 2*q + 1, q prime    *
*************************************************/
bool DL_Group::prime_is_safe() const
   {
   if(type() == Uninitialized)
      throw Invalid_State("DLP group cannot be used uninitialized");
   if(type() == DH_Group)
      return true;
   else
      return false;
   }

/*************************************************
* DER encode the paramaters                      *
*************************************************/
SecureVector<byte> DL_Group::DER_encode() const
   {
   if(type() == Uninitialized)
      throw Invalid_State("DLP group cannot be used uninitialized");

   DER_Encoder encoder;
   encoder.start_sequence();
   if(type() == DSA_Group)
      {
      DER::encode(encoder, p);
      DER::encode(encoder, q);
      DER::encode(encoder, g);
      }
   else if(type() == DH_Group)
      {
      DER::encode(encoder, p);
      DER::encode(encoder, g);
      DER::encode(encoder, q);
      }
   else
      throw Invalid_Argument("Unknown DL group type " + to_string(type()));
   encoder.end_sequence();

   return encoder.get_contents();
   }

/*************************************************
* PEM encode the paramaters                      *
*************************************************/
std::string DL_Group::PEM_encode() const
   {
   SecureVector<byte> encoding = DER_encode();
   if(type() == DH_Group)
      return pem_encode(encoding, encoding.size(), "DH PARAMETERS");
   else
      return pem_encode(encoding, encoding.size(), "DSA PARAMETERS");
   }

/*************************************************
* Decode BER encoded paramaters                  *
*************************************************/
void DL_Group::BER_decode(DataSource& source, GroupType type)
   {
   BigInt new_p, new_q, new_g;

   BER_Decoder decoder(source);
   BER_Decoder sequence = BER::get_subsequence(decoder);
   if(type == DSA_Group)
      {
      BER::decode(sequence, new_p);
      BER::decode(sequence, new_q);
      BER::decode(sequence, new_g);
      }
   else if(type == DH_Group)
      {
      BER::decode(sequence, new_p);
      BER::decode(sequence, new_g);
      BER::decode(sequence, new_q);
      sequence.discard_remaining();
      }
   else
      throw Invalid_Argument("Unknown DL group type " + to_string(type));
   sequence.verify_end();

   p = new_p;
   q = new_q;
   g = new_g;
   group_type = type;
   if(!fast_check())
      throw Invalid_Argument("DL_Group: Invalid primes have been specified");
   }

/*************************************************
* Decode PEM encoded paramaters                  *
*************************************************/
void DL_Group::PEM_decode(DataSource& source)
   {
   Pipe ber;
   std::string label;
   ber.process_msg(pem_decode(source, label));

   if(label == "DH PARAMETERS")
      BER_decode(ber, DH_Group);
   else if(label == "DSA PARAMETERS")
      BER_decode(ber, DSA_Group);
   else
      throw Decoding_Error("DL_Group: Invalid PEM label " + label);
   }

}
