/*************************************************
* PKCS PBE Base Source File                      *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/pbe_pkcs.h>
#include <botan/lookup.h>
#include <botan/rng.h>
#include <memory>

namespace Botan {

/*************************************************
* Encrypt some bytes using a PBE                 *
*************************************************/
void PBE_PKCS::write(const byte input[], u32bit length)
   {
   static const u32bit BUFFERSIZE = DEFAULT_BUFFERSIZE;
   while(length)
      {
      u32bit put = std::min(BUFFERSIZE, length);
      pipe.write(input, length);
      flush_pipe(true);
      length -= put;
      }
   }

/*************************************************
* Start encrypting with a PBE                    *
*************************************************/
void PBE_PKCS::start_msg()
   {
   if(!key_set)
      throw Invalid_State(name() + ": No key has been set");
   cipher_obj->reset_key(key);
   cipher_obj->reset_iv(iv);

   pipe.start_msg();
   if(pipe.message_count() > 1)
      pipe.set_default_msg(pipe.default_msg() + 1);
   }

/*************************************************
* Finish encrypting with a PBE                   *
*************************************************/
void PBE_PKCS::end_msg()
   {
   pipe.end_msg();
   flush_pipe(false);
   }

/*************************************************
* Flush the pipe                                 *
*************************************************/
void PBE_PKCS::flush_pipe(bool safe_to_skip)
   {
   static const u32bit BUFFERSIZE = DEFAULT_BUFFERSIZE;

   if(safe_to_skip && pipe.remaining() < 64)
      return;

   SecureBuffer<byte, BUFFERSIZE> buffer;
   while(pipe.remaining())
      {
      u32bit got = pipe.read(buffer, BUFFERSIZE);
      send(buffer, got);
      }
   }

/*************************************************
* Set the passphrase to use                      *
*************************************************/
void PBE_PKCS::set_key(const std::string& passphrase)
   {
   const u32bit BLOCK_SIZE = block_size_of(cipher);
   const u32bit KEY_LENGTH = keylength_for(cipher);
   const u32bit DERIVE = KEY_LENGTH + BLOCK_SIZE;
   std::auto_ptr<S2K> pbkdf(get_s2k(pbkdf_name() + "(" + digest + ")"));
   pbkdf->set_iterations(iterations);
   pbkdf->change_salt(salt, salt.size());
   SymmetricKey key_and_iv = pbkdf->derive_key(DERIVE, passphrase);

   key.set(key_and_iv, KEY_LENGTH);
   iv.set(key_and_iv + KEY_LENGTH, BLOCK_SIZE);
   key_set = true;
   }

/*************************************************
* Create a new set of generic PBE parameters     *
*************************************************/
void PBE_PKCS::new_params()
   {
   iterations = 65536;
   salt.create(8);
   Global_RNG::randomize(salt, salt.size(), Nonce);
   key_set = false;
   }

/*************************************************
* PKCS PBE Base Constructor                      *
*************************************************/
PBE_PKCS::PBE_PKCS(const std::string& d_algo, const std::string& c_algo,
                   Cipher_Dir dir)
   {
   digest = deref_alias(d_algo);
   std::vector<std::string> cipher_spec = split_on_slash(c_algo);
   if(cipher_spec.size() != 2)
      throw Invalid_Argument("PBE-PKCS: Invalid cipher spec " + c_algo);
   cipher = deref_alias(cipher_spec[0]);
   const std::string cipher_mode = cipher_spec[1];

   if(!have_block_cipher(cipher))
      throw Algorithm_Not_Found(cipher);
   if(!have_hash(digest))
      throw Algorithm_Not_Found(digest);
   key_set = false;

   const u32bit BLOCK_SIZE = block_size_of(cipher);
   const u32bit KEY_LENGTH = max_keylength_of(cipher);
   SymmetricKey dummy_key(BLOCK_SIZE);
   InitializationVector dummy_iv(KEY_LENGTH);
   cipher_obj = get_cipher(c_algo, dummy_key, dummy_iv, dir);
   pipe.append(cipher_obj);
   }

}
