/*************************************************
* PKCS #5 PBES2 Source File                      *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/pbe_pkcs.h>
#include <botan/lookup.h>
#include <botan/rng.h>
#include <botan/asn1.h>
#include <memory>

namespace Botan {

/*************************************************
* Encrypt some bytes using PBES2                 *
*************************************************/
void PBE_PKCS5v20::write(const byte input[], u32bit length)
   {
   static const u32bit BUFFERSIZE = DEFAULT_BUFFERSIZE;
   while(length)
      {
      u32bit put = std::min(BUFFERSIZE, length);
      pipe.write(input, length);
      flush_pipe(true);
      length -= put;
      }
   }

/*************************************************
* Start encrypting with PBES2                    *
*************************************************/
void PBE_PKCS5v20::start_msg()
   {
   if(!key_set)
      throw Invalid_State("PBE-PKCS5 v2.0: No key has been set");
   cipher_obj->reset_key(key);
   cipher_obj->reset_iv(iv);

   pipe.start_msg();
   if(pipe.message_count() > 1)
      pipe.set_default_msg(pipe.default_msg() + 1);
   }

/*************************************************
* Finish encrypting with PBES2                   *
*************************************************/
void PBE_PKCS5v20::end_msg()
   {
   pipe.end_msg();
   flush_pipe(false);
   }

/*************************************************
* Flush the pipe                                 *
*************************************************/
void PBE_PKCS5v20::flush_pipe(bool safe_to_skip)
   {
   static const u32bit BUFFERSIZE = DEFAULT_BUFFERSIZE;

   if(safe_to_skip && pipe.remaining() < 64)
      return;

   SecureBuffer<byte, BUFFERSIZE> buffer;
   while(pipe.remaining())
      {
      u32bit got = pipe.read(buffer, BUFFERSIZE);
      send(buffer, got);
      }
   }

/*************************************************
* Set the passphrase to use                      *
*************************************************/
void PBE_PKCS5v20::set_key(const std::string& passphrase)
   {
   const u32bit DERIVE = block_size_of(cipher) + key_length;
   std::auto_ptr<S2K> pbkdf(get_s2k("PBKDF2(" + digest + ")"));
   pbkdf->set_iterations(iterations);
   pbkdf->change_salt(salt, salt.size());
   SymmetricKey key_and_iv = pbkdf->derive_key(DERIVE, passphrase);

   key.set(key_and_iv, key_length);
   iv.set(key_and_iv + key_length, block_size_of(cipher));
   key_set = true;
   }

/*************************************************
* Create a new set of PBES2 parameters           *
*************************************************/
void PBE_PKCS5v20::new_params()
   {
   iterations = 65536;
   key_length = max_keylength_of(cipher);
   salt.create(8);
   Global_RNG::randomize(salt, salt.size(), Nonce);
   key_set = false;
   }

/*************************************************
* Encode PKCS#5 PBES2 parameters                 *
*************************************************/
SecureVector<byte> PBE_PKCS5v20::encode_params() const
   {
   AlgorithmIdentifier kdf_algo, enc_algo;

   DER_Encoder encoder;
   encoder.start_sequence();
   DER::encode(encoder, kdf_algo);
   DER::encode(encoder, enc_algo);
   encoder.end_sequence();
   return encoder.get_contents();
   }

/*************************************************
* Decode PKCS#5 PBES2 parameters                 *
*************************************************/
void PBE_PKCS5v20::decode_params(DataSource& source)
   {
   AlgorithmIdentifier kdf_algo, enc_algo;

   BER_Decoder decoder(source);
   BER_Decoder sequence = BER::get_subsequence(decoder);
   BER::decode(sequence, kdf_algo);
   BER::decode(sequence, enc_algo);
   sequence.verify_end();

   key_set = false;
   }

/*************************************************
* Return an OID for this PBES2 type              *
*************************************************/
OID PBE_PKCS5v20::get_oid() const
   {
   return OID("1.2.840.113549.1.5.13");
   }

/*************************************************
* PKCS#5 v2.0 PBE Constructor                    *
*************************************************/
PBE_PKCS5v20::PBE_PKCS5v20(const std::string& d_algo,
                           const std::string& c_algo,
                           Cipher_Dir dir) : direction(dir)
   {
   digest = deref_alias(d_algo);
   std::vector<std::string> cipher_spec = split_on_slash(c_algo);
   if(cipher_spec.size() != 2)
      throw Invalid_Argument("PBE-PKCS5 v2.0: Invalid cipher spec " + c_algo);
   cipher = deref_alias(cipher_spec[0]);
   const std::string cipher_mode = cipher_spec[1];

   if(!have_block_cipher(cipher))
      throw Algorithm_Not_Found(cipher);
   if(!have_hash(digest))
      throw Algorithm_Not_Found(digest);
   if((cipher != "DES" && cipher != "TripleDES") || (cipher_mode != "CBC"))
      throw Invalid_Argument("PBE-PKCS5 v2.0: Invalid cipher " + c_algo);
   if(digest != "SHA-160")
      throw Invalid_Argument("PBE-PKCS5 v2.0: Invalid digest " + digest);
   key_set = false;

   SymmetricKey dummy_key(8);
   InitializationVector dummy_iv(8);
   cipher_obj = get_cipher(c_algo, dummy_key, dummy_iv, dir);
   pipe.append(cipher_obj);
   }

}
